;;; org-multiline-macro.el --- Define block macros with #+begin_macro  -*- lexical-binding: t; -*-

;; Copyright (C) 2024 Umut Tuna Akgül

;; This program is free software: you can redistribute it and or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation, either version 3 of the
;; License, or (at your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see
;; <https://www.gnu.org/licenses/>.

;;; Commentary:
;; This package allows defining multiline macros via #+BEGIN_MACRO
;; #END_MACRO blocks.  The syntax of such a block is simply:
;; 
;; #+BEGIN_MACRO MACRO-NAME [PARAMETERS...]
;; ...
;; #+END_MACRO
;;
;; MACRO-NAME is any sequence of non-whitespace characters, and it is
;; equivalent to the way it is used in
;;
;; #+MACRO: MACRO-NAME BODY...
;;
;; PARAMETERS is in the same format as it is in source blocks, but the
;; only meaningful parameter is currently `:eval'.
;;
;; If the `:eval' parameter is "yes" or the body of the macro starts
;; with "(eval", the macro is taken to be emacs lisp code, and the
;; macro is defined the way it is via
;;
;; #+MACRO: FOO (eval BODY)
;;
;; Otherwise, the body of the macro is what the macro expands to, the
;; same way it is when defined via
;;
;; #+MACRO: FOO BODY...

;;; TODOs:
;; - Have emacs lisp macro blocks be fontified and editable the same
;;   way emacs lisp source blocks are.
;; - Maybe define a minor mode to turn macro blocks on and off?  I
;;   don't think it's a good idea but who knows.
;; - Make ":eval no" work even with an "(eval" macro.

(require 'advice)
(require 'org-element)
(require 'ob-core)

(defun org-multiline-macro--collect-1 (element)
  "Return a cons cell of the form (NAME . MACRO-BODY) from ELEMENT.
If ELEMENT is not a valid macro block, return nil."
  (when-let (((org-element-type-p element 'special-block))
             (type (org-element-property :type element))
             ((string-equal-ignore-case "macro" type)))
    (let* ((parameters (org-element-property :parameters element))
           (name
            (prog2
                (or (string-match "^\\s-*\\(\\S-+\\)" parameters)
                    (error "Macro name missing"))
                (match-string 1 parameters)
              (setq parameters
                    (org-babel-parse-header-arguments
                     (substring parameters (match-end 1))
                     t))))
           (eval (cdr (assq 'eval parameters)))
           (contents
            (string-trim
             (with-current-buffer (org-element-property :buffer element)
               (without-restriction
                 (buffer-substring (org-element-property :contents-begin element)
                                   (org-element-property :contents-end element)))))))
      (cons name
            (if (and eval (string-equal-ignore-case "yes" eval)
                     (not (string-match-p "\\`(eval\\>" contents)))
                (concat "(eval (progn " contents "\n))")
              contents)))))

(defun org-multiline-macro--collect ()
  "Collect block macro definitions in the current buffer and setup files.
Return an alist containing all macro templates found."
  (append
   (org-element-map (org-element-parse-buffer 'element nil t) 'special-block
     #'org-multiline-macro--collect-1)
   (org-multiline-macro--collect-setupfiles)))

;; This was pulled from `org--collect-keywords-1' and lightly
;; modified.
(defun org-multiline-macro--collect-setupfiles ()
  "Collect block macro definitions in the SETUPFILES of the current buffer.
This macro recursively descends down SETUPFILES, skipping any that have
already been visited."
  (let (alist)
    (save-excursion
      (without-restriction
        (goto-char (point-min))
        (let ((case-fold-search t)
              (regexp (org-make-options-regexp '("SETUPFILE"))))
          (while (re-search-forward regexp nil t)
            (when-let
                ((element (org-element-at-point))
                 ((org-element-type-p element 'keyword))
                 ((string-equal-ignore-case
                   (org-element-property :key element)
                   "SETUPFILE"))
                 (value (org-element-property :value element)))
              (setq alist (append alist (org-multiline-macro--collect-uri value))))))))
    alist))

(defvar org-multiline-macro--visiting-files nil)

(defun org-multiline-macro--collect-uri (uri)
  (when (org-string-nw-p uri)
    (let* ((uri (org-strip-quotes uri))
           (uri-is-url (org-url-p uri))
           (uri (if uri-is-url uri
                  ;; In case of error, be safe.
                  ;; See bug#68976.
                  (ignore-errors    ; return nil when expansion fails.
		    (expand-file-name uri)))))
      (when (and uri (not (member uri org-multiline-macro--visiting-files)))
        (let ((org-multiline-macro--visiting-files
               (cons uri org-multiline-macro--visiting-files)))
          (with-temp-buffer
	    (unless uri-is-url
	      (setq default-directory (file-name-directory uri)))
	    (let ((contents (org-file-contents uri :noerror)))
	      (when contents
	        (insert contents)
                ;; CHECK: Do we still not need to setup org mode?
	        (let ((major-mode 'org-mode))
		  (org-multiline-macro--collect))))))))))

(define-advice org-macro--collect-macros
    (:filter-return (templates) collect-multiline)
  (append (org-multiline-macro--collect) templates))

(provide 'org-multiline-macro)

;;; org-multiline-macro.el ends here
